@@ -1,6 +1,6 @@
 ;;; Consfigurator -- Lisp declarative configuration management system
 
-;;; Copyright (C) 2021  Sean Whitton <spwhitton@spwhitton.name>
+;;; Copyright (C) 2021-2022  Sean Whitton <spwhitton@spwhitton.name>
 
 ;;; This file is free software; you can redistribute it and/or modify
 ;;; it under the terms of the GNU General Public License as published by
@@ -46,27 +46,23 @@
 
 (defvar *fork-control* nil)
 
-(defmacro with-fork-request (prerequest request (out err exit) &body forms)
-  (with-gensyms (input output)
-    `(progn
-       (unless (lisp-connection-p)
-         (failed-change "Forking requires a Lisp-type connection."))
-       (unless *fork-control*
-         (failed-change
-          "Fork requested but no fork control child; is this the root Lisp?"))
-       (informat 3 "~&Making grandchild request ~S" ,request)
-       (with-mkfifos (,input ,output)
-         ;; We send the path to a named pipe, INPUT, rather than our actual
-         ;; request.  That way we can be confident that what we send into the
-         ;; (shared) requests pipe will be less than PIPE_BUF (see pipe(7)).
-         (write-to-mkfifo (cons ,input ,output) *fork-control*)
-         (with-open-file (,input ,input :direction :output :if-exists :append
-                                        :element-type 'character)
-           (write-to-mkfifo ,prerequest ,input)
-           (write-to-mkfifo ,request    ,input))
-         (destructuring-bind (,out ,err ,exit)
-             (safe-read-file-form ,output :element-type 'character)
-           ,@forms)))))
+(defun issue-fork-request (prerequest request)
+  (unless (lisp-connection-p)
+    (failed-change "Forking requires a Lisp-type connection."))
+  (unless *fork-control*
+    (failed-change
+     "Fork requested but no fork control child; is this the root Lisp?"))
+  (informat 3 "~&Making grandchild request ~S" request)
+  (with-mkfifos (input output)
+    ;; We send the path to a named pipe, INPUT, rather than our actual
+    ;; request.  That way we can be confident that what we send into the
+    ;; (shared) requests pipe will be less than PIPE_BUF (see pipe(7)).
+    (write-to-mkfifo (cons input output) *fork-control*)
+    (with-open-file (input input :direction :output :if-exists :append
+                           :element-type 'character)
+      (write-to-mkfifo prerequest input)
+      (write-to-mkfifo request    input))
+    (values-list (safe-read-file-form output :element-type 'character))))
 
 ;;; These are the two requests we expect to make of grandchildren: complete
 ;;; the work of an enclosing call to DEPLOY* or DEPLOY-THESE*, or dump an
@@ -75,32 +71,29 @@
 ;;; the latter case we do not carry over any of these by default.
 
 (defun wrap-grandchild-request (&rest forms)
-  ``(let ((*host* ,*host*)
-          (*connection* ,*connection*)
-          (*no-data-sources* t)
-          (*consfigurator-debug-level* ,*consfigurator-debug-level*))
-      ,,@forms))
+  `(let ((*host* ,*host*)
+         (*connection* ,*connection*)
+         (*no-data-sources* t)
+         (*consfigurator-debug-level* ,*consfigurator-debug-level*))
+     ,@forms))
 
-(defmacro eval-in-grandchild (prerequest request (out err exit) &body forms)
+(defun eval-in-grandchild (prerequest request)
   "Evaluate PREREQUEST and REQUEST, both readably printable Lisp forms, in a
 grandchild process.  PREREQUEST and REQUEST must be evaluable using only
 definitions established statically by your consfig, or in one of the ASDF
-systems upon which your consfig depends.  Then bind OUT, ERR and EXIT to the
-stdout, stderr and exit code of that process, respectively, and evaluate
-FORMS.
+systems upon which your consfig depends.  Returns the stdout, stderr and exit
+code of that process.
 
 PREREQUEST will be evaluated before the grandchild calls fork(2) to establish
 its own infrastructure for subsequent uses of this macro, and REQUEST after.
 Thus, PREREQUEST must not start up any threads."
-  `(with-fork-request
-       ,(wrap-grandchild-request
-         '`(posix-login-environment
-            ,(get-connattr :remote-uid)
-            ,(get-connattr :remote-user)
-            ,(get-connattr :remote-home))
-         prerequest)
-       ,(wrap-grandchild-request request) (,out ,err ,exit)
-     ,@forms))
+  (issue-fork-request
+   (wrap-grandchild-request `(posix-login-environment
+                              ,(get-connattr :remote-uid)
+                              ,(get-connattr :remote-user)
+                              ,(get-connattr :remote-home))
+                            prerequest)
+   (wrap-grandchild-request request)))
 
 ;; As SBCL does not expose the build-id, use two checksums.
 #+sbcl (defvar *sbcl-core-cksum*) #+sbcl (defvar *sbcl-runtime-cksum*)
@@ -125,7 +118,20 @@ Thus, PREREQUEST must not start up any threads."
 
 (define-simple-error wrong-execution-context-for-image-dump (aborted-change))
 
-(defun %dump-consfigurator-in-grandchild (filename pre-dump form)
+(defun dump-consfigurator-in-grandchild (filename pre-dump form)
+  "Dump an executable image to FILENAME which will evaluate the readably
+printable Lisp form FORM, which defaults to one which will execute the current
+deployment.  FORM must be evaluable using only definitions established
+statically by your consfig, or in one of the ASDF systems upon which your
+consfig depends.  Evaluate PRE-DUMP in the process which will perform the dump
+prior to dumping.
+
+Only :LISP property :APPLY subroutines should call this.
+
+The process which performs the dump will have its umask set to #o077, but
+implementation-specific image dumping code might undo this (SBCL, for example,
+changes the mode of the file to #o755).  You might want to ensure that the
+directory containing FILENAME is locked down."
   ;; Check that the image is likely to be reinvokable.
   (loop for library in (list-foreign-libraries)
         for path = (foreign-library-pathname library)
@@ -151,72 +157,46 @@ Thus, PREREQUEST must not start up any threads."
     (wrong-execution-context-for-image-dump
      "Couldn't dump executable image because same SBCL build unavailable."))
   ;; Perform the image dump.
-  (with-fork-request nil
-      `(progn ,pre-dump
-              (nix:umask #o077)
-              (uiop:register-image-restore-hook
-               (named-lambda hook ()
-                 ;; Remove ourselves so that a dump-and-reinvoke by the
-                 ;; reinvoked image will not try to run us again.
-                 (deletef uiop:*image-restore-hook* #'hook)
-                 (eval ',form))
-               nil)
-              (uiop:dump-image ,filename :executable t))
-      (out err exit)
+  (multiple-value-bind (out err exit)
+      (issue-fork-request
+       nil
+       `(progn ,pre-dump
+               (nix:umask #o077)
+               (uiop:register-image-restore-hook
+                (named-lambda hook ()
+                  ;; Remove ourselves so that a dump-and-reinvoke by the
+                  ;; reinvoked image will not try to run us again.
+                  (deletef uiop:*image-restore-hook* #'hook)
+                  (eval ',form))
+                nil)
+               (uiop:dump-image ,filename :executable t)))
     (declare (ignore out))
     (unless (zerop exit)
       (failed-change "~&Failed to dump image; stderr was ~%~%~A" err))))
 
-(defmacro eval-in-reinvoked (prerequest request (out err exit) &body forms)
+(defun eval-in-reinvoked (prerequest request)
   "In a grandchild process, evaluate PREREQUEST, dump an executable image, and
 immediately reinvoke that image to evaluate REQUEST.  PREREQUEST and REQUEST
 must be evaluable using only definitions established statically by your
-consfig, or in one of the ASDF systems upon which your consfig depends.  Then
-bind OUT, ERR and EXIT to the stdout, stderr and exit code of that process,
-respectively, and evaluate FORMS."
-  (with-gensyms (tempdir)
-    ;; Create a temporary directory which will be readable only by the present
-    ;; user, because of how SBCL overrides the umask when dumping an image.
-    ;; Don't want to use ~/.cache/consfigurator/images because want to write
-    ;; to a tmpfs/ramdisk if possible.
-    `(with-local-temporary-directory (,tempdir)
-       (let ((file (merge-pathnames "consfigurator" ,tempdir)))
-         (%dump-consfigurator-in-grandchild
-          file ,(wrap-grandchild-request prerequest)
-          ;; Try to ensure that the new fork control child does not end up
-          ;; with the actual request in its memory.
-          '(with-backtrace-and-exit-code
-            (with-fork-control (eval (with-standard-io-syntax (read))))))
-         (nix:chmod file #o700)         ; ensure it's executable
-         (multiple-value-bind (,out ,err ,exit)
-             (run :may-fail :input (with-standard-io-syntax
-                                     (write-to-string
-                                      ,(wrap-grandchild-request request)))
-                  file)
-           ,@forms)))))
-
-(defun dump-consfigurator-in-grandchild
-    (filename &optional (form `(let ((*no-data-sources* t)
-                                     (*connection* ,*connection*)
-                                     (*consfigurator-debug-level*
-                                       ,*consfigurator-debug-level*))
-                                 (with-deployment-report
-                                   (with-fork-control
-                                     (%consfigure nil ,*host*)))
-                                 (fresh-line))))
-  "Dump an executable image to FILENAME which will evaluate the readably
-printable Lisp form FORM, which defaults to one which will execute the current
-deployment.  FORM must be evaluable using only definitions established
-statically by your consfig, or in one of the ASDF systems upon which your
-consfig depends.
-
-Only :LISP property :APPLY subroutines should call this.
-
-The process which performs the dump will have its umask set to #o077, but
-implementation-specific image dumping code might undo this (SBCL, for example,
-changes the mode of the file to #o755).  You might want to ensure that the
-directory containing FILENAME is locked down."
-  (%dump-consfigurator-in-grandchild filename nil form))
+consfig, or in one of the ASDF systems upon which your consfig depends.
+Returns the stdout, stderr and exit code of that process."
+  ;; Create a temporary directory which will be readable only by the present
+  ;; user, because of how SBCL overrides the umask when dumping an image.
+  ;; Don't want to use ~/.cache/consfigurator/images because want to write to
+  ;; a tmpfs/ramdisk if possible.
+  (with-local-temporary-directory (tempdir)
+    (let ((file (merge-pathnames "consfigurator" tempdir)))
+      (dump-consfigurator-in-grandchild
+       file (wrap-grandchild-request prerequest)
+       ;; Try to ensure that the new fork control child does not end up with
+       ;; the actual request in its memory.
+       '(with-backtrace-and-exit-code
+         (with-fork-control (eval (with-standard-io-syntax (read))))))
+      (nix:chmod file #o700)            ; ensure it's executable
+      (run :may-fail :input (with-standard-io-syntax
+                              (write-to-string
+                               (wrap-grandchild-request request)))
+           file))))
 
 (defprop image-dumped :lisp (&optional filename form (always form))
   "Dump an executable image to FILENAME which will evaluate FORM, which must be
@@ -283,9 +263,14 @@ property by applying it like this:
                 ;; should not just be quietly skipped over.
                 (apply #'aborted-change (simple-condition-format-control error)
                        (simple-condition-format-arguments error))))))
-       (if form
-           (dump-consfigurator-in-grandchild file form)
-           (dump-consfigurator-in-grandchild file))))
+       (dump-consfigurator-in-grandchild
+        file nil (or form `(let ((*no-data-sources* t)
+                                 (*connection* ,*connection*)
+                                 (*consfigurator-debug-level*
+                                   ,*consfigurator-debug-level*))
+                             (with-deployment-report
+                               (with-fork-control (%consfigure nil ,*host*)))
+                             (fresh-line))))))
    ;; Return :NO-CHANGE, though we can't detect whether a change was actually
    ;; made: it depends on whether the definitions determining the evaluation
    ;; of FORM, or the definition of this host established by the consfig, was
@@ -303,7 +288,8 @@ property by applying it like this:
          ;; standard about closing synonym streams; see
          ;; <https://bugs.launchpad.net/sbcl/+bug/1904257>.
          (loop initially (mapc-open-input-streams
-                          #'close *standard-input* *debug-io* *terminal-io*)
+                          #'close (list *standard-input*
+                                        *debug-io* *terminal-io*))
                with ,fork-control = (open ,fork-control
                                           :element-type 'character)
                for (input . output) = (handler-case (with-safe-io-syntax ()
@@ -312,8 +298,8 @@ property by applying it like this:
                                           (close ,fork-control)
                                           (uiop:quit)))
                do (mapc-open-output-streams
-                   #'force-output
-                   *standard-output* *error-output* *debug-io* *terminal-io*)
+                   #'force-output (list *debug-io* *terminal-io*
+                                        *standard-output* *error-output*))
                when (zerop (fork))
                  do (nix:setsid)
                     (close ,fork-control)
@@ -447,8 +433,7 @@ Called by connection types which start up remote Lisp images."
                   :goal-operation 'asdf:monolithic-compile-bundle-op))
           (let ((name (asdf:component-name requirement)))
             ;; Handle UIOP specially because it comes with ASDF.
-            (unless
-                (memstring= (asdf:primary-system-name name) '("asdf" "uiop"))
+            (unless (memstr= (asdf:primary-system-name name) '("asdf" "uiop"))
               (pushnew requirement asdf-requirements)))))
       (nreversef asdf-requirements))
     requirements))
@@ -620,5 +605,5 @@ hostattrs; see \"Pitfalls\" in the Consfigurator user manual.
 
 If ~:*~A is a simple object then you may be able to resolve this by defining
 a PRINT-OBJECT method for your class, possibly using
-CONSFIGURATOR:DEFINE-PRINT-OBJECT-FOR-STRUCTLIKE."
+CONSFIGURATOR:DEFINE-SIMPLE-PRINT-OBJECT."
                    (print-not-readable-object c))))))))
