@@ -1,6 +1,6 @@
 ;;; Consfigurator -- Lisp declarative configuration management system
 
-;;; Copyright (C) 2021  Sean Whitton <spwhitton@spwhitton.name>
+;;; Copyright (C) 2021-2022  Sean Whitton <spwhitton@spwhitton.name>
 
 ;;; This file is free software; you can redistribute it and/or modify
 ;;; it under the terms of the GNU General Public License as published by
@@ -611,6 +611,11 @@ possible.  Ignored if VOLUME-SIZE is also bound."))
     "The value of the --type parameter to cryptsetup luksFormat.
 Note that GRUB2 older than 2.06 cannot open the default LUKS2 format, so
 specify \"luks1\" if this is needed.")
+   (cryptsetup-options
+    :type list :initform nil :initarg :cryptsetup-options
+    :documentation
+    "Extra arguments to pass to cryptsetup(8) when creating the volume, such as
+'--cipher'.  Use the LUKS-TYPE slot for '--type'.")
    (crypttab-options
     :type list :initform '("luks" "discard" "initramfs")
     :initarg :crypttab-options :accessor crypttab-options)
@@ -632,12 +637,15 @@ specify \"luks1\" if this is needed.")
                         (merge-pathnames volume-label #P"/dev/mapper/"))))
 
 (defmethod create-volume ((volume luks-container) (file pathname))
-  (with-slots (luks-passphrase-iden1 volume-label luks-type) volume
+  (with-slots
+        (luks-passphrase-iden1 volume-label luks-type cryptsetup-options)
+      volume
     (mrun :inform
           :input (get-data-string luks-passphrase-iden1 (volume-label volume))
           "cryptsetup" "--type" luks-type
           (and (member luks-type '("luks" "luks2") :test #'string=)
                `("--label" ,volume-label))
+          cryptsetup-options
           "luksFormat" file "-")))
 
 (defmethod close-volume ((volume opened-luks-container))
@@ -667,12 +675,16 @@ order in which they should be closed, and as a second value, a corresponding
 list of the immediate parents of each opened volume.  MOUNT-BELOW specifies a
 pathname to prefix to mount points when opening FILESYSTEM volumes.
 
+Also return as third and fourth values a list of volumes encountered that were
+already open and a corresponding list of their immediate parents.
+
 Calling this function can be useful for testing at the REPL, but code should
 normally use WITH-OPEN-VOLUMES or WITH-OPENED-VOLUMES.
 
 If an error is signalled while the attempt to open volumes is in progress, a
 single attempt will be made to close all volumes opened up to that point."
-  (let (opened-volumes all-parents filesystems)
+  (let
+      (opened-volumes opened-parents already-open already-parents filesystems)
     (handler-case
         (labels
             ((open-volume-and-contents (volume file parent)
@@ -690,14 +702,19 @@ single attempt will be made to close all volumes opened up to that point."
                        (if (subtypep (class-of volume) 'opened-volume)
                            (setq opened-volumes
                                  (append opened-contents opened-volumes)
-                                 all-parents
-                                 (nconc opened-contents-parents all-parents))
+                                 opened-parents
+                                 (nconc opened-contents-parents
+                                        opened-parents)
+                                 already-open
+                                 (cons volume already-open)
+                                 already-parents
+                                 (cons parent already-parents))
                            (setq opened-volumes
                                  (append opened-contents
                                          (cons opened opened-volumes))
-                                 all-parents
+                                 opened-parents
                                  (nconc opened-contents-parents
-                                        (cons parent all-parents)))))
+                                        (cons parent opened-parents)))))
                      (dolist (opened-volume (or opened-contents `(,opened)))
                        (when (slot-boundp opened-volume 'volume-contents)
                          (open-volume-and-contents
@@ -713,9 +730,9 @@ single attempt will be made to close all volumes opened up to that point."
                               :key (compose #'ensure-directory-pathname
                                             #'mount-point
                                             #'cadr))))
-                (push (pop filesystem) all-parents)
+                (push (pop filesystem) opened-parents)
                 (push (apply #'open-volume filesystem) opened-volumes)))
-          (values opened-volumes all-parents))
+          (values opened-volumes opened-parents already-open already-parents))
       (serious-condition (condition)
         (unwind-protect (mapc #'close-volume opened-volumes)
           (error condition))))))
@@ -743,24 +760,26 @@ populate /etc/fstab and /etc/crypttab.  Do not modify this list."
          (mapc #'close-volume ,opened-volumes)))))
 
 (defmacro with-opened-volumes
-    ((volumes &key (mount-below nil mount-below-supplied-p)) &body propapps)
+    ((volumes &key (mount-below nil mount-below-supplied-p) leave-open)
+     &body propapps)
   "Macro property combinator.  Where each of VOLUMES is a VOLUME which may be
 opened by calling OPEN-VOLUME with NIL as the second argument, recursively
-open each of VOLUMES and any contents thereof, apply PROPAPPS, and close all
-volumes that were opened.
+open each of VOLUMES and any contents thereof, apply PROPAPPS, and, unless
+LEAVE-OPEN, close all volumes that were opened.
 
 MOUNT-BELOW specifies a pathname to prefix to mount points when opening
 FILESYSTEM volumes.  During the application of PROPAPPS, all
 'DISK:OPENED-VOLUMES and 'DISK:OPENED-VOLUME-PARENTS connattrs are replaced
-with lists of the volumes that were opened and corresponding immediate parent
-volumes; the former must not be modified."
+with lists of the volumes that were opened/already open and corresponding
+immediate parent volumes."
   `(with-opened-volumes*
      ',volumes
      ,(if (cdr propapps) `(eseqprops ,@propapps) (car propapps))
-     ,@(and mount-below-supplied-p `(:mount-below ,mount-below))))
+     ,@(and mount-below-supplied-p `(:mount-below ,mount-below))
+     :leave-open ,leave-open))
 
 (define-function-property-combinator with-opened-volumes*
-    (volumes propapp &key (mount-below nil mount-below-supplied-p))
+    (volumes propapp &key (mount-below nil mount-below-supplied-p) leave-open)
   (:retprop
    :type (propapp-type propapp)
    :hostattrs (lambda-ignoring-args
@@ -768,15 +787,17 @@ volumes; the former must not be modified."
                 (propapp-attrs propapp))
    :apply
    (lambda-ignoring-args
-     (multiple-value-bind (opened-volumes parents)
+     (multiple-value-bind
+           (opened-volumes opened-parents already-open already-parents)
          (apply #'open-volumes-and-contents
                 `(,volumes ,@(and mount-below-supplied-p
                                   `(:mount-below ,mount-below))))
-       (with-connattrs ('opened-volumes opened-volumes
-                        'opened-volume-parents parents)
+       (with-connattrs
+           ('opened-volumes (append opened-volumes already-open)
+            'opened-volume-parents (append opened-parents already-parents))
          (unwind-protect (apply-propapp propapp)
            (mrun "sync")
-           (mapc #'close-volume (get-connattr 'opened-volumes))))))
+           (unless leave-open (mapc #'close-volume opened-volumes))))))
    :args (cdr propapp)))
 
 (defun create-volumes-and-contents (volumes &optional files)
@@ -907,6 +928,21 @@ the LVM physical volumes corresponding to those volume groups."
   (ensure-directory-pathname
    (strcat (unix-namestring image-pathname) ".chroot")))
 
+(defun host-volumes-just-one-physical-disk (host fun)
+  (loop
+    with found
+    for volume in (get-hostattrs :volumes (preprocess-host host))
+    for physical-disk-p = (subtypep (type-of volume) 'physical-disk)
+    if (and physical-disk-p (not found) (slot-boundp volume 'volume-contents))
+      do (setq found t)
+      and collect (aprog1 (copy-volume-and-contents volume) (funcall fun it))
+    else unless physical-disk-p
+           collect volume
+    finally
+       (unless found
+         (inapplicable-property
+          "Volumes list for host has no DISK:PHYSICAL-DISK with contents."))))
+
 (defpropspec raw-image-built-for :lisp
     (options host image-pathname &key rebuild)
   "Build a raw disk image for HOST at IMAGE-PATHNAME.
@@ -933,30 +969,117 @@ of the host's volumes changes, although the contents of the image's
 filesystems will be incrementally updated when other properties change."
   (:desc #?"Built image for ${(get-hostname host)} @ ${image-pathname}")
   (let ((chroot (image-chroot image-pathname))
-        (volumes
-          (loop
-            with found
-            for volume in (get-hostattrs :volumes (preprocess-host host))
-            for physical-disk-p = (subtypep (type-of volume) 'physical-disk)
-            if (and physical-disk-p (not found)
-                    (slot-boundp volume 'volume-contents))
-              do (setq found t)
-              and collect (aprog1 (copy-volume-and-contents volume)
-                            (change-class it 'raw-disk-image)
-                            (setf (image-file it) image-pathname))
-            else unless physical-disk-p
-                   collect volume
-            finally
-               (unless found
-                 (inapplicable-property
-                  "Volumes list for host has no DISK:PHYSICAL-DISK with contents.")))))
+        (volumes (host-volumes-just-one-physical-disk
+                  host (lambda (volume)
+                         (change-class volume 'raw-disk-image)
+                         (setf (image-file volume) image-pathname)))))
     `(on-change (eseqprops
                  ,(propapp (chroot:os-bootstrapped-for. options chroot host
                              (caches-cleaned)))
                  (%raw-image-created
                   ,volumes :chroot ,chroot :rebuild ,rebuild))
-       (consfigurator.property.installer:chroot-installed-to-volumes-for
-        ,host ,chroot ,volumes))))
+       (consfigurator.property.installer:files-installed-to-volumes-for
+        nil ,host ,volumes :chroot ,chroot))))
+
+(defprop %volumes-created :lisp (volumes)
+  (:desc "Created host volumes")
+  (:hostattrs (os:required 'os:linux) (require-volumes-data volumes))
+  (:apply
+   (dolist (volume volumes)
+     (when (subtypep (type-of volume) 'physical-disk)
+       (setf (volume-size volume)
+             (floor (/ (parse-integer
+                        (stripln
+                         (run "blockdev" "--getsize64" (device-file volume))))
+                       1048576)))))
+   (create-volumes-and-contents volumes)))
+
+(defpropspec first-disk-installed-for :lisp
+    (options host device-file &key chroot)
+  "Install HOST to the DISK:PHYSICAL-DISK accessible at DEVICE-FILE.
+**THIS PROPERTY UNCONDITIONALLY FORMATS DISKS, POTENTIALLY DESTROYING DATA,
+  EACH TIME IT IS APPLIED.**
+
+Do not apply in DEFHOST.  Apply with DEPLOY-THESE/HOSTDEPLOY-THESE.
+
+The DISK:VOLUME-CONTENTS of the first DISK:PHYSICAL-DISK entry in the host's
+volumes, as specified using DISK:HAS-VOLUMES, will be created at DEVICE-FILE;
+there must be at least one such DISK:PHYSICAL-DISK entry.  Other
+DISK:PHYSICAL-DISK entries will be ignored, so ensure that none of the
+properties of the host will write to areas of the filesystem where filesystems
+stored on other physical disks would normally be mounted.
+
+OPTIONS will be passed on to CHROOT:OS-BOOTSTRAPPED-FOR, which see.
+
+In most cases you will need to ensure that HOST has properties which do at
+least the following:
+
+  - declare the host's OS
+
+  - install a kernel
+
+  - install the binaries/packages needed to install the host's bootloader to
+    its volumes (usually with INSTALLER:BOOTLOADER-BINARIES-INSTALLED).
+
+If CHROOT, an intermediate chroot is bootstrapped at CHROOT, and HOST's
+properties are applied to that.  Otherwise, HOST's OS is bootstrapped directly
+to DEVICE-FILE.  It's useful to supply CHROOT when you expect to install the
+same HOST to a number of physical disks.
+
+Applying this property is similar to applying DISK:RAW-IMAGE-BUILT-FOR and
+then immediately dd'ing out the image to DEVICE-FILE.  The advantage of this
+property is that there is no need to resize filesystems to fill the size of
+the host's actual physical disk upon first boot."
+  (:desc #?"Installed ${(get-hostname host)} to ${device-file}")
+  (let ((volumes (host-volumes-just-one-physical-disk
+                  host (lambda (v)
+                         (setf (device-file v)
+                               (parse-unix-namestring device-file))))))
+    `(eseqprops
+      (%volumes-created ,volumes)
+      ,@(and chroot
+             (list (propapp (chroot:os-bootstrapped-for. options chroot host
+                              (caches-cleaned)))))
+      (consfigurator.property.installer:files-installed-to-volumes-for
+       ,options ,host ,volumes :chroot ,chroot))))
+
+(defpropspec volumes-installed-for :lisp (options host &key chroot leave-open)
+  "Install HOST to its volumes, as specified using DISK:HAS-VOLUMES.
+**THIS PROPERTY UNCONDITIONALLY FORMATS DISKS, POTENTIALLY DESTROYING DATA,
+  EACH TIME IT IS APPLIED.**
+
+Do not apply in DEFHOST.  Apply with DEPLOY-THESE/HOSTDEPLOY-THESE.
+
+OPTIONS will be passed on to CHROOT:OS-BOOTSTRAPPED-FOR, which see.
+
+In most cases you will need to ensure that HOST has properties which do at
+least the following:
+
+  - declare the host's OS
+
+  - install a kernel
+
+  - install the binaries/packages needed to install the host's bootloader to
+    its volumes (usually with INSTALLER:BOOTLOADER-BINARIES-INSTALLED).
+
+If CHROOT, an intermediate chroot is bootstrapped at CHROOT, and HOST's
+properties are applied to that.  Otherwise, HOST's OS is bootstrapped directly
+to its volumes.  This parameter is useful for the case of installing HOST from
+a live system which might not have network access.  See \"Tutorial: OS
+installation\" in the Consfigurator user's manual.
+
+If LEAVE-OPEN, HOST's volumes will not be closed.  This allows you to inspect
+the result of the installation.  If you want to run this property again, you
+should first manually close all the volumes."
+  (:desc #?"Installed ${(get-hostname host)}")
+  (let ((volumes (get-hostattrs :volumes host)))
+    `(eseqprops
+      (%volumes-created ,volumes)
+      ,@(and chroot
+             (list (propapp (chroot:os-bootstrapped-for. options chroot host
+                              (caches-cleaned)))))
+      (consfigurator.property.installer:files-installed-to-volumes-for
+       ,options ,host ,volumes :chroot ,chroot :leave-open ,leave-open))))
 
 (defprop %squashfsed :posix (chroot image &optional (compression "xz"))
   (:apply
@@ -1063,15 +1186,6 @@ Currently only BIOS boot is implemented."
 
          ,iso-root)))))
 
-(defprop host-volumes-created :lisp ()
-  "Recursively create the volumes as specified by DISK:HAS-VOLUMES.
-**THIS PROPERTY UNCONDITIONALLY FORMATS DISKS, POTENTIALLY DESTROYING DATA,
-  EACH TIME IT IS APPLIED.**
-
-Do not apply in DEFHOST.  Apply with DEPLOY-THESE/HOSTDEPLOY-THESE."
-  (:desc "Host volumes created")
-  (:apply (create-volumes-and-contents (get-hostattrs :volumes))))
-
 ;; TODO Possibly we want (a version of) this to not fail, but just do nothing,
 ;; if the relevant volume groups etc. are inactive?
 (defproplist host-logical-volumes-exist :lisp ()
